static char help[] = "Tests repeated VecSetType().\n\n";

#include <petscvec.h>
#include <petscmanagedtype.hpp>
#include <petsc/private/deviceimpl.h>
#include <petscdevice_cupm.h>
#include <cuda_profiler_api.h>

int main(int argc, char **argv)
{
  using namespace Petsc;
  Vec x;
  PetscDeviceContext dctx;

  PetscFunctionBeginUser;
  PetscCall(PetscInitialize(&argc, &argv, (char *)0, help));

  /* create vector */
  PetscCall(VecCreate(PETSC_COMM_WORLD, &x));
  PetscCall(VecSetSizes(x, 100, PETSC_DECIDE));
  PetscCall(VecSetType(x, VECCUDA));
  PetscCall(VecSetFromOptions(x));
  PetscCall(PetscDeviceContextGetCurrentContext(&dctx));
  {
    ManagedScalar scal;
    PetscScalar *device_ptr;
    MPI_Comm scomm;
    PetscScalar begin, end;

    PetscCall(scal.Reserve(dctx, 1));
    scal.at(dctx, 0) = 1.0;
    PetscCall(VecSetRandom(x, nullptr));
    PetscCall(VecNormAsync(x, NORM_2, &scal, dctx));
    PetscCallCUDA(cudaMalloc(&device_ptr, sizeof(*device_ptr)));
    PetscUseTypeMethod(dctx, getstreamcomm, PETSC_COMM_WORLD, &scomm);
    PetscCallCUDA(cudaDeviceSynchronize());
    PetscCallCUDA(cudaProfilerStart());
    PetscCall(PetscTime(&begin));
    for (PetscInt i = 0; i < 100; ++i) {
      #if 0
      PetscCall(VecNormAsync(x, NORM_2, &scal, dctx));
      #else
      //PetscCall(VecScaleAsync(x, scal, dctx));
      PetscCallMPI(MPIX_Allreduce_enqueue(MPI_IN_PLACE, device_ptr, 1, MPIU_SCALAR, MPIU_SUM, scomm));
      #endif
    }
    PetscCall(PetscTime(&end));
    PetscCallCUDA(cudaProfilerStop());
    PetscCallCUDA(cudaFree(device_ptr));
    PetscCall(PetscPrintf(PETSC_COMM_WORLD, "time %g\n", (double)(end - begin)));
  }

  PetscCall(VecDestroy(&x));

  PetscCall(PetscFinalize());
  return 0;
}
